/*
	Kuwahara plugin for AviSynth --  The (generalized) Kuwahara filter 
	is an edge-preserving smoothing filter. The algorithm is described
	in http://en.wikipedia.org/wiki/Kuwahara_filter (and its references).

	Copyright (C) 2014 Wilbert Dijkhof
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at:
	wilbertdijkhof at yahoo.com
*/

#include <windows.h>
#include <math.h>       /* pow */
#include "avisynth.h"

class Kuwahara : public GenericVideoFilter {
	int window;
	bool generalized;
	int q;
	bool smooth_chroma;
	int edge;
	int threshold;
	double edge_brightness;
	VideoInfo vi_luma;
public:
   Kuwahara(PClip _child, int _window, bool _generalized, int _q, bool _smooth_chroma, int _edge, double _threshold, double _edge_brightness, IScriptEnvironment* env);
   void Luma(PVideoFrame &src, PVideoFrame &src_luma, IScriptEnvironment* env);
   void CalculateStatistics(PVideoFrame &src_luma, int window, int x, int y, double &m_local, double &var_local);
   void CalculateAverage(PVideoFrame &src, int window, int x, int y, int channel, double &m_local);
   PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

Kuwahara::Kuwahara(PClip _child, int _window, bool _generalized, int _q, bool _smooth_chroma, int _edge, double _threshold, double _edge_brightness, IScriptEnvironment* env) :
   GenericVideoFilter(_child), window(_window), generalized(_generalized), q(_q), smooth_chroma(_smooth_chroma), edge(_edge), threshold(_threshold), edge_brightness(_edge_brightness) {

   if (!vi.IsRGB()) {
      env->ThrowError("Kuwahara: RGB data only!");
   }

   if (!(edge_brightness>=0 && edge_brightness<=255)) {
	   env->ThrowError("Kuwahara: edge_brightness must be between 0 and 255!");
   }

   if (!(edge==0 || edge==1 || edge==2)) {
	   env->ThrowError("Kuwahara: possible values for edge are 0, 1 or 2!");
   }

   memset(&vi_luma, 0, sizeof(VideoInfo));
   vi_luma = vi;
   vi_luma.pixel_type = VideoInfo::CS_Y8;
}

PVideoFrame __stdcall Kuwahara::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);

	PVideoFrame src_luma = env->NewVideoFrame(vi_luma);
	Luma(src, src_luma, env);

	int src_pitch = src->GetPitch();
	const unsigned char* srcp = src->GetReadPtr();
	int src_width = src->GetRowSize();
	int src_height = src->GetHeight();
	int bpp = vi.BytesFromPixels(1);

	int dst_pitch = dst->GetPitch();
	unsigned char* dstp = dst->GetWritePtr();
	int dst_width = dst->GetRowSize();
	int dst_height = dst->GetHeight();

	int src_luma_pitch = src_luma->GetPitch();

	int x, y, quadrant;
	double m, m1, var1, m2, var2, m3, var3, m4, var4, var_grey, var_min;
	double m_blue, m_green, m_red, m1_blue, m1_green, m1_red, m2_blue, m2_green, m2_red, m3_blue, m3_green, m3_red, m4_blue, m4_green, m4_red;
	int luma;

	double m_local = 0;
	double var_local = 0;

	// copy edges
	if (edge==1) {
		memset(dstp, 255, dst_pitch * window);
	} else {
		env->BitBlt(dstp, dst_pitch, srcp, src_pitch, src_width, window);
	}
	srcp += src_pitch * window;
	dstp += dst_pitch * window;

	for (y=window; y<dst_height-window; y++) {
		if (edge==1) {
			memset(dstp, 255, window*bpp);
			memset(dstp+dst_width-window*bpp, 255, window*bpp);
		} else {
			env->BitBlt(dstp, window*bpp, srcp, window*bpp, window*bpp, 1);
			env->BitBlt(dstp+dst_width-window*bpp, window*bpp, srcp+dst_width-window*bpp, window*bpp, window*bpp, 1);
		}
		srcp += src_pitch;
		dstp += dst_pitch;
	}

	if (edge==1) {
		memset(dstp, 255, dst_pitch * window);
	} else {
		env->BitBlt(dstp, dst_pitch, srcp, src_pitch, src_width, window);
	}

	srcp = src->GetReadPtr() + src_pitch * window;
	dstp = dst->GetWritePtr() + dst_pitch * window;

	for (y=window; y<dst_height-window; y++) {
		for (x=window*bpp; x<dst_width-window*bpp; x+=bpp) {

			const unsigned char* srcp_luma = src_luma->GetReadPtr() + src_luma_pitch * (vi_luma.height-1-y);
			luma = srcp_luma[int(x/bpp)];

			//calculate mean (m) and variance (var) of the luma pixels in every quadrant (w = window) for pixel (x,y)
			//quadrant I - [x,x+w] x [y,y+w] ( (w+1)^2 pixels )

			CalculateStatistics(src_luma, window, x, y, m_local, var_local);

			m1 = m_local;
			var1 = var_local;

			if (edge>0) {
				CalculateStatistics(src_luma, 1, x, y, m_local, var_local);
				var_grey = var_local;
			} else { // not used
				var_grey = 0;
			}

			//quadrant II - [x-w,x] x [y,y+w] ( (w+1)^2 pixels )
			CalculateStatistics(src_luma, window, x-window*bpp, y, m_local, var_local);

			m2 = m_local;
			var2 = var_local;

			//quadrant III - [x-w,x] x [y-w,y] ( (w+1)^2 pixels )
			CalculateStatistics(src_luma, window, x-window*bpp, y-window, m_local, var_local);

			m3 = m_local;
			var3 = var_local;

			//quadrant IV - [x,x+w] x [y-w,y] ( (w+1)^2 pixels )
			CalculateStatistics(src_luma, window, x, y-window, m_local, var_local);

			m4 = m_local;
			var4 = var_local;
			
			// selection (m,var)
			quadrant = 0;
			if (!generalized) { // standard Kuwahara filter
				quadrant = 1;
				var_min = var1;
				if (var2<var_min) {
					quadrant = 2;
					var_min = var2;
				}
				if (var3<var_min) {
					quadrant = 3;
					var_min = var3;
				}
				if (var4<var_min) {
					quadrant = 4;
					var_min = var4;
				}
			}

			if (!generalized) { // standard Kuwahara filter
				switch (quadrant) {
				case 1:
					if (smooth_chroma) {
						CalculateAverage(src, window, x, y, 0, m_blue);
						CalculateAverage(src, window, x, y, 1, m_green);
						CalculateAverage(src, window, x, y, 2, m_red);
					} else {
						m = m1;
					}
					break;
				case 2:
					if (smooth_chroma) {
						CalculateAverage(src, window, x-window*bpp, y, 0, m_blue);
						CalculateAverage(src, window, x-window*bpp, y, 1, m_green);
						CalculateAverage(src, window, x-window*bpp, y, 2, m_red);
					} else {
						m = m2;
					}
					break;
				case 3:
					if (smooth_chroma) {
						CalculateAverage(src, window, x-window*bpp, y-window, 0, m_blue);
						CalculateAverage(src, window, x-window*bpp, y-window, 1, m_green);
						CalculateAverage(src, window, x-window*bpp, y-window, 2, m_red);
					} else {
						m = m3;
					}
					break;
				case 4:
					if (smooth_chroma) {
						CalculateAverage(src, window, x, y-window, 0, m_blue);
						CalculateAverage(src, window, x, y-window, 1, m_green);
						CalculateAverage(src, window, x, y-window, 2, m_red);
					} else {
						m = m4;
					}
					break;
				}
			} else { // generalized variant of the Kuwahara filter (q = input parameter)
				// phi_q = sum_i (m_i / s_i^q) / sum_i (1 / s_i^q)  (s_i = sqrt(vari))
				// except if vari<0.001 then m=mi.
				double var_sum_inv;
				if (var1<0.001) {
					if (smooth_chroma) {
						CalculateAverage(src, window, x, y, 0, m_blue);
						CalculateAverage(src, window, x, y, 1, m_green);
						CalculateAverage(src, window, x, y, 2, m_red);
					} else {
						m = m1;
					}
				} else if (var2<0.001) {
					if (smooth_chroma) {
						CalculateAverage(src, window, x-window*bpp, y, 0, m_blue);
						CalculateAverage(src, window, x-window*bpp, y, 1, m_green);
						CalculateAverage(src, window, x-window*bpp, y, 2, m_red);
					} else {
						m = m2;
					}
				} else if (var3<0.001) {
					if (smooth_chroma) {
						CalculateAverage(src, window, x-window*bpp, y-window, 0, m_blue);
						CalculateAverage(src, window, x-window*bpp, y-window, 1, m_green);
						CalculateAverage(src, window, x-window*bpp, y-window, 2, m_red);
					} else {
						m = m3;
					}
				} else if (var4<0.001) {
					if (smooth_chroma) {
						CalculateAverage(src, window, x, y-window, 0, m_blue);
						CalculateAverage(src, window, x, y-window, 1, m_green);
						CalculateAverage(src, window, x, y-window, 2, m_red);
					} else {
						m = m4;
					}
				} else {
					var_sum_inv = pow(1/var1,q) + pow(1/var2,q) + pow(1/var3,q) + pow(1/var4,q);
					if (smooth_chroma) {
						CalculateAverage(src, window, x, y, 0, m1_blue);
						CalculateAverage(src, window, x, y, 1, m1_green);
						CalculateAverage(src, window, x, y, 2, m1_red);
						CalculateAverage(src, window, x-window*bpp, y, 0, m2_blue);
						CalculateAverage(src, window, x-window*bpp, y, 1, m2_green);
						CalculateAverage(src, window, x-window*bpp, y, 2, m2_red);
						CalculateAverage(src, window, x-window*bpp, y-window, 0, m3_blue);
						CalculateAverage(src, window, x-window*bpp, y-window, 1, m3_green);
						CalculateAverage(src, window, x-window*bpp, y-window, 2, m3_red);
						CalculateAverage(src, window, x, y-window, 0, m4_blue);
						CalculateAverage(src, window, x, y-window, 1, m4_green);
						CalculateAverage(src, window, x, y-window, 2, m4_red);
						m_blue = (m1_blue / pow(var1,q) + m2_blue / pow(var2,q) + m3_blue / pow(var3,q) + m4_blue / pow(var4,q)) / var_sum_inv;
						m_green = (m1_green / pow(var1,q) + m2_green / pow(var2,q) + m3_green / pow(var3,q) + m4_green / pow(var4,q)) / var_sum_inv;
						m_red = (m1_red / pow(var1,q) + m2_red / pow(var2,q) + m3_red / pow(var3,q) + m4_red / pow(var4,q)) / var_sum_inv;
					} else {
						m = (m1 / pow(var1,q) + m2 / pow(var2,q) + m3 / pow(var3,q) + m4 / pow(var4,q)) / var_sum_inv;
					}
				}
			}

			if ((var1>threshold) && (var2>threshold) && (var3>threshold) && (var4>threshold)) { // don't smooth if surrounded by edges
				if ((edge==0) || ((edge==2) && !(var_grey>threshold))) {
					dstp[x] = srcp[x];
					dstp[x+1] = srcp[x+1];
					dstp[x+2] = srcp[x+2];
				}
			} else {
				if ((edge==0) || (edge==2)) {
					if (smooth_chroma) {
						dstp[x] = min(255,max(0,int(m_blue + 0.5)));
						dstp[x+1] = min(255,max(0,int(m_green + 0.5)));
						dstp[x+2] = min(255,max(0,int(m_red + 0.5)));
					} else {
						dstp[x] = min(255,max(0,int(srcp[x] - luma + m + 0.5)));
						dstp[x+1] = min(255,max(0,int(srcp[x+1] - luma + m + 0.5)));
						dstp[x+2] = min(255,max(0,int(srcp[x+2] - luma + m + 0.5)));
					}
				}
			}

			if (((edge==1) || (edge==2)) && (var_grey>threshold)) {
				dstp[x] = min(255,max(0,int(edge_brightness*srcp[x] + 0.5)));
				dstp[x+1] = min(255,max(0,int(edge_brightness*srcp[x+1] + 0.5)));
				dstp[x+2] = min(255,max(0,int(edge_brightness*srcp[x+2] + 0.5)));
			} else if ((edge==1) && !(var_grey>threshold)) {
				dstp[x] = 255;
				dstp[x+1] = 255;
				dstp[x+2] = 255;
			}

			if (bpp==4) { // copy alpha if present
				dstp[x+3] = srcp[x+3];
			}

		}
		srcp += src_pitch;
		dstp += dst_pitch;
	}

	return dst;

}

AVSValue __cdecl Create_Kuwahara(AVSValue args, void* user_data, IScriptEnvironment* env) {
   return new Kuwahara(args[0].AsClip(),
					   args[1].AsInt(1),			// window
					   args[2].AsBool(true),		// generalized
					   args[3].AsInt(2),			// q
					   args[4].AsBool(false),		// smooth_chroma
					   args[5].AsInt(0),			// edge (0 = no edge, 1 = edge mask, 2 = draws edge on clip)
					   args[6].AsFloat(70),			// threshold for edge detection
					   args[7].AsFloat(0.75),		// edge_brightness
					   env);
}

const AVS_Linkage *AVS_linkage = 0;

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit3(IScriptEnvironment* env, const AVS_Linkage* const vectors) {
   AVS_linkage = vectors;
   env->AddFunction("Kuwahara", "c[window]i[generalized]b[q]i[smooth_chroma]b[edge]i[threshold]f[edge_brightness]f", Create_Kuwahara, 0);
   return "Kuwahara plugin";
}

void Kuwahara::Luma(PVideoFrame &src, PVideoFrame &src_luma, IScriptEnvironment* env)
{
	const double kr = 0.299;
	const double kg = 0.587;
	const double kb = 0.114;
	int h, w;
	double y;

	int src_pitch = src->GetPitch();
	const unsigned char* srcp = src->GetReadPtr() + src_pitch * (vi.height-1);  // We start at last line (RGB is upside down)
	int src_width = src->GetRowSize();
	int src_height = src->GetHeight();

	unsigned char* srcp_luma = src_luma->GetWritePtr();
	int src_luma_pitch = src_luma->GetPitch();
	int src_luma_width = src_luma->GetRowSize();
	int src_luma_height = src_luma->GetHeight();

	int bpp = vi.BytesFromPixels(1);

	for (h=0; h<src_luma_height; h++) {
		for (w=0; w<src_luma_width; w++) {
			//calculate unscaled luma (PC.601)
			y = kr*srcp[2]+kg*srcp[1]+kb*srcp[0];
			srcp_luma[w] = (int)(y+0.5);
			srcp += bpp;
		}
		srcp -= (src_pitch+vi.width*bpp);
		srcp_luma += src_luma_pitch;
	}
}

void Kuwahara::CalculateStatistics(PVideoFrame &src_luma, int window, int x, int y, double &m_local, double &var_local)
{
	//calculate mean (m) and variance (var) of the luma pixels in every quadrant for samples (x,y) <-> pixel (x/bpp,y)
	//quadrant I - [x,x+window] x [y,y+window] ( (window+1)^2 pixels )

	int bpp = vi.BytesFromPixels(1);
	int h, w;
	int src_luma_pitch = src_luma->GetPitch();
	const unsigned char* srcp_luma = src_luma->GetReadPtr() + src_luma_pitch * (vi_luma.height-1-y);

	m_local=0;
	var_local=0;

	for (h=0; h<window+1; h++) {
		for (w=0; w<window+1; w++) {
			m_local = m_local + srcp_luma[int(x/bpp)+w];
		}
		srcp_luma -= src_luma_pitch;
	}
	m_local = m_local/((window+1)*(window+1));

	srcp_luma = src_luma->GetReadPtr() + src_luma_pitch * (vi_luma.height-1-y);
	for (h=0; h<window+1; h++) {
		for (w=0; w<window+1; w++) {
			var_local = var_local + (srcp_luma[int(x/bpp)+w]-m_local)*(srcp_luma[int(x/bpp)+w]-m_local);
		}
		srcp_luma -= src_luma_pitch;
	}
	var_local = var_local/((window+1)*(window+1));
}

void Kuwahara::CalculateAverage(PVideoFrame &src, int window, int x, int y, int channel, double &m_local)
{
	//calculate mean (m) of the color samples (channel) in every quadrant for sample (x,y) [thus sample, not pixel]
	//channel: 0 (blue), 1 (green) or 2 (red)
	//quadrant I - [x,x+window] x [y,y+window] ( (window+1)^2 pixels )

	int bpp = vi.BytesFromPixels(1);
	int h, w;
	int src_pitch = src->GetPitch();
	const unsigned char* srcp = src->GetReadPtr() + src_pitch * y;

	m_local=0;

	for (h=0; h<window+1; h++) {
		for (w=0; w<window+1; w++) {
			m_local = m_local + srcp[x+channel+w*bpp];
		}
		srcp += src_pitch;
	}
	m_local = m_local/((window+1)*(window+1));
}
